@@ -1,6 +1,6 @@
 /***
  * ASM: a very small and fast Java bytecode manipulation framework
- * Copyright (c) 2000-2005 INRIA, France Telecom
+ * Copyright (c) 2000-2007 INRIA, France Telecom
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -111,7 +111,7 @@ public class MethodNode extends MemberNode implements MethodVisitor {
      * @associates org.objectweb.asm.tree.AbstractInsnNode
      * @label instructions
      */
-    public List instructions;
+    public InsnList instructions;
 
     /**
      * The try catch blocks of this method. This list is a list of
@@ -140,13 +140,12 @@ public class MethodNode extends MemberNode implements MethodVisitor {
     public List localVariables;
 
     /**
-     * The line numbers of this method. This list is a list of
-     * {@link LineNumberNode} objects. May be <tt>null</tt>
-     * 
-     * @associates org.objectweb.asm.tree.LineNumberNode
+     * Constructs an unitialized {@link MethodNode}.
      */
-    public List lineNumbers;
-
+    public MethodNode() {
+        this.instructions = new InsnList();
+    }
+    
     /**
      * Constructs a new {@link MethodNode}.
      * 
@@ -167,6 +166,7 @@ public class MethodNode extends MemberNode implements MethodVisitor {
         final String signature,
         final String[] exceptions)
     {
+        this();
         this.access = access;
         this.name = name;
         this.desc = desc;
@@ -175,10 +175,8 @@ public class MethodNode extends MemberNode implements MethodVisitor {
                 ? 0
                 : exceptions.length);
         boolean isAbstract = (access & Opcodes.ACC_ABSTRACT) != 0;
-        this.instructions = new ArrayList(isAbstract ? 0 : 24);
         if (!isAbstract) {
             this.localVariables = new ArrayList(5);
-            this.lineNumbers = new ArrayList(5);
         }
         this.tryCatchBlocks = new ArrayList();
         if (exceptions != null) {
@@ -192,7 +190,7 @@ public class MethodNode extends MemberNode implements MethodVisitor {
 
     public AnnotationVisitor visitAnnotationDefault() {
         return new AnnotationNode(new ArrayList(0) {
-            public boolean add(Object o) {
+            public boolean add(final Object o) {
                 annotationDefault = o;
                 return super.add(o);
             }
@@ -230,6 +228,20 @@ public class MethodNode extends MemberNode implements MethodVisitor {
     public void visitCode() {
     }
 
+    public void visitFrame(
+        final int type,
+        final int nLocal,
+        final Object[] local,
+        final int nStack,
+        final Object[] stack)
+    {
+        instructions.add(new FrameNode(type, nLocal, local == null
+                ? null
+                : getLabelNodes(local), nStack, stack == null
+                ? null
+                : getLabelNodes(stack)));
+    }
+
     public void visitInsn(final int opcode) {
         instructions.add(new InsnNode(opcode));
     }
@@ -242,8 +254,8 @@ public class MethodNode extends MemberNode implements MethodVisitor {
         instructions.add(new VarInsnNode(opcode, var));
     }
 
-    public void visitTypeInsn(final int opcode, final String desc) {
-        instructions.add(new TypeInsnNode(opcode, desc));
+    public void visitTypeInsn(final int opcode, final String type) {
+        instructions.add(new TypeInsnNode(opcode, type));
     }
 
     public void visitFieldInsn(
@@ -265,11 +277,11 @@ public class MethodNode extends MemberNode implements MethodVisitor {
     }
 
     public void visitJumpInsn(final int opcode, final Label label) {
-        instructions.add(new JumpInsnNode(opcode, label));
+        instructions.add(new JumpInsnNode(opcode, getLabelNode(label)));
     }
 
     public void visitLabel(final Label label) {
-        instructions.add(new LabelNode(label));
+        instructions.add(getLabelNode(label));
     }
 
     public void visitLdcInsn(final Object cst) {
@@ -286,7 +298,10 @@ public class MethodNode extends MemberNode implements MethodVisitor {
         final Label dflt,
         final Label[] labels)
     {
-        instructions.add(new TableSwitchInsnNode(min, max, dflt, labels));
+        instructions.add(new TableSwitchInsnNode(min,
+                max,
+                getLabelNode(dflt),
+                getLabelNodes(labels)));
     }
 
     public void visitLookupSwitchInsn(
@@ -294,7 +309,9 @@ public class MethodNode extends MemberNode implements MethodVisitor {
         final int[] keys,
         final Label[] labels)
     {
-        instructions.add(new LookupSwitchInsnNode(dflt, keys, labels));
+        instructions.add(new LookupSwitchInsnNode(getLabelNode(dflt),
+                keys,
+                getLabelNodes(labels)));
     }
 
     public void visitMultiANewArrayInsn(final String desc, final int dims) {
@@ -307,7 +324,10 @@ public class MethodNode extends MemberNode implements MethodVisitor {
         final Label handler,
         final String type)
     {
-        tryCatchBlocks.add(new TryCatchBlockNode(start, end, handler, type));
+        tryCatchBlocks.add(new TryCatchBlockNode(getLabelNode(start),
+                getLabelNode(end),
+                getLabelNode(handler),
+                type));
     }
 
     public void visitLocalVariable(
@@ -321,13 +341,13 @@ public class MethodNode extends MemberNode implements MethodVisitor {
         localVariables.add(new LocalVariableNode(name,
                 desc,
                 signature,
-                start,
-                end,
+                getLabelNode(start),
+                getLabelNode(end),
                 index));
     }
 
     public void visitLineNumber(final int line, final Label start) {
-        lineNumbers.add(new LineNumberNode(line, start));
+        instructions.add(new LineNumberNode(line, getLabelNode(start)));
     }
 
     public void visitMaxs(final int maxStack, final int maxLocals) {
@@ -335,6 +355,42 @@ public class MethodNode extends MemberNode implements MethodVisitor {
         this.maxLocals = maxLocals;
     }
 
+    /**
+     * Returns the LabelNode corresponding to the given Label. Creates a new 
+     * LabelNode if necessary. The default implementation of this method uses
+     * the {@link Label#info} field to store associations between labels and
+     * label nodes.
+     * 
+     * @param l a Label.
+     * @return the LabelNode corresponding to l.
+     */    
+    protected LabelNode getLabelNode(final Label l) {
+        if (!(l.info instanceof LabelNode)) {
+            l.info = new LabelNode(l);
+        }
+        return (LabelNode) l.info;
+    }
+
+    private LabelNode[] getLabelNodes(final Label[] l) {
+        LabelNode[] nodes = new LabelNode[l.length];
+        for (int i = 0; i < l.length; ++i) {
+            nodes[i] = getLabelNode(l[i]);
+        }
+        return nodes;
+    }
+
+    private Object[] getLabelNodes(final Object[] objs) {
+        Object[] nodes = new Object[objs.length];
+        for (int i = 0; i < objs.length; ++i) {
+            Object o = objs[i];
+            if (o instanceof Label) {
+                o = getLabelNode((Label) o);
+            }
+            nodes[i] = o;
+        }
+        return nodes;
+    }
+
     // ------------------------------------------------------------------------
     // Accept method
     // ------------------------------------------------------------------------
@@ -356,7 +412,7 @@ public class MethodNode extends MemberNode implements MethodVisitor {
             accept(mv);
         }
     }
-    
+
     /**
      * Makes the given method visitor visit this method.
      * 
@@ -368,7 +424,9 @@ public class MethodNode extends MemberNode implements MethodVisitor {
         if (annotationDefault != null) {
             AnnotationVisitor av = mv.visitAnnotationDefault();
             AnnotationNode.accept(av, null, annotationDefault);
-            av.visitEnd();
+            if (av != null) {
+                av.visitEnd();
+            }
         }
         n = visibleAnnotations == null ? 0 : visibleAnnotations.size();
         for (i = 0; i < n; ++i) {
@@ -418,19 +476,12 @@ public class MethodNode extends MemberNode implements MethodVisitor {
                 ((TryCatchBlockNode) tryCatchBlocks.get(i)).accept(mv);
             }
             // visits instructions
-            for (i = 0; i < instructions.size(); ++i) {
-                ((AbstractInsnNode) instructions.get(i)).accept(mv);
-            }
+            instructions.accept(mv);
             // visits local variables
             n = localVariables == null ? 0 : localVariables.size();
             for (i = 0; i < n; ++i) {
                 ((LocalVariableNode) localVariables.get(i)).accept(mv);
             }
-            // visits line numbers
-            n = lineNumbers == null ? 0 : lineNumbers.size();
-            for (i = 0; i < n; ++i) {
-                ((LineNumberNode) lineNumbers.get(i)).accept(mv);
-            }
             // visits maxs
             mv.visitMaxs(maxStack, maxLocals);
         }
